﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace ProcessRubyScript
{
    public class RubyExpression
    {
        private string _getExprScript;
        private string _getTruncScript;
        private string _rubyExe;
        private Process _getTruncProcess;

        public RubyExpression()
        {
            //Below works at work on MY computer only.
            //_rubyExe = @"c:\Ruby186\ruby\bin\rubyw.exe";

            /*
             * Below did not work at work on MY computer: c:/ruby/lib/ruby/1.8/i386-mingw32/rbconfig.rb is
             * at 1.8.6 but c:\ruby\bin\rubyw.exe is ver. 1.8.7. [This line does work at home however, both
             * are 1.8.6.] USE THIS LINE FOR DCS TESTING AT WORK AND FOR RELEASING TO PRODUCTION, those
             * computers are at 1.8.6 for both c:/ruby/lib... and c:\ruby\bin\rubyw.exe
            */
            _rubyExe = @"c:\ruby\bin\rubyw.exe";

            SetupGetTruncProcess();

            /*
             * Using a verbatim string (@) so we can 'see' the Ruby script better. Note that the
             * double quote must now be doubled to be escaped, but everything else is as is.
             * 
             * Used when we are checking for truncation, in the process we need to check if we have
             * a complete expression. We do this to check the old style of truncation where the 
             * truncation code would be placed at the end (i.e. when it is complete) of the assignment
             * expression. We also check for truncation on the print line, this is where the Dataflow 
             * Retriever puts it.
            */
            _getTruncScript = @"
                require 'tk'

                def CompleteExpression?(str, b)

                    old_verbose = $VERBOSE
                    $VERBOSE = nil

                    catch(:valid) do
                        eval(""BEGIN{throw :valid}\n#{str}"", b)
                    end

                    # Non comment lines that end in a comma or backslash are not complete.
                    (str =~ /^\s*#/ ? true : str !~ /[,\\]\s*\z/)

                    rescue SyntaxError => error
                        false
                    ensure
                        $VERBOSE = old_verbose
                end

                def get_binding(param)
                    return binding
                end

                def CheckTruncation(str, varName, varHash)

	                return if (str.strip =~ /\s*=\s*''$/)

	                truncStatus = (str.strip =~ /\[\s*0\s*,\s*(\d+)\s*\]$/ ? $1 : ""NT"")

	                if (varHash.has_key?(varName))

		                varHash[varName] = ""DIFF"" unless (varHash[varName] == truncStatus)
	                else
		                varHash[varName] = truncStatus
	                end
                end

                clipStr = TkClipboard.get(nil)
                varHash = Hash.new()
                printHash = Hash.new()
                varName = """"
                truncLength = """"
                compExpStr = """";
                chkCompExp = false;

                strBinding = get_binding(clipStr)

                clipStr.each do |line|

	                if (chkCompExp)

		                compExpStr += line

		                if (CompleteExpression?(compExpStr, strBinding))

			                CheckTruncation(line, varName, varHash)
			                varName = """"
			                chkCompExp = false
		                end
	                elsif (line =~ /^\s*([a-z][a-z0-9_]+)\s*=\s*.+/)
			
		                varName = $1

		                if (CompleteExpression?(line, strBinding))
			
			                CheckTruncation(line, varName, varHash)
			                varName = """"
		                else
			                compExpStr += line
			                chkCompExp = true
		                end
	                elsif (line =~ /\s*print\s*'(box\d+_\d+)'\s*,\s*([a-z][a-z0-9_]+)\s*(\[\s*0\s*,\s*(\d+)\s*\])?/)

		                if (varHash.has_key?($2))

			                if ($4)
				                if (varHash[$2] == ""DIFF"")
					                truncLength = ""DIFF""
				                elsif (varHash[$2] == ""NT"")
					                truncLength = $4
				                else
					                truncLength = (varHash[$2] == $4 ? $4 : ""DIFF"")
				                end
			                else
				                truncLength = varHash[$2]
			                end

			                varHash.delete($2)
		                else
			                truncLength = ($4 ? $4 : ""NT"")
		                end

		                if (printHash.has_key?($1))

			                printHash[$1] = ""DIFF"" if (printHash[$1] != truncLength)
		                else
			                printHash[$1] = truncLength
		                end
	                end
                end

                printHash.each { |key, value|
	                unless (value == ""NT"")
		                print ""#{key}|#{value}\x03""
	                end
                }";

            /*
             * Return syntactically 'complete' Ruby expressions. We also perform a conservative
             * identification of non-snippet conditional expression blocks, like
             * 'if (!individual.married_joint.strip.empty?)' for example, which starts a block.
             * 
             * We do this by running a control string which is checked for a complete expression at each
             * line read from the clipboard. This process is started when we discover one of the non-snippet
             * conditionals. At the same time, the conditional code logic runs it's own conditional string,
             * looking for complete sub-expressions which are inside the conditional block(s) [they can be
             * nested] - we also handle 'else's. When we read the last line which will complete the control
             * string expression we check if the conditional code also identifies the expression as complete.
             * We also check if the conditional code concludes a complete expression while the control string
             * is still not complete.
             * 
             * Three checks are performed in this regard - at each line read one of these MAY be true:
             * 
             * 1) The control string is not complete but the conditional string code says it is, this is
             *      an early completion. We dump the conditional info and just wait for the control string
             *      to be complete at which time we print it as one complete expression.
             * 2) The control string is complete but the conditional code does not agree. The conditional
             *      code is somehow out of synch with the control string, so we dump the conditional info
             *      and print the control string as one complete expression.
             * 3) The control string is complete AND the conditional code also agrees with the completion
             *      - this is the hopefull outcome, for this we rely on the condCompAck flag which is the
             *      acknowledgment of the control string assertion. We print the conditional string as a
             *      series of complete sub-expressions which comprise the control string expression.
            */
            _getExprScript = @"
                require 'tk'

                def CompleteExpression?(str, b)

                    old_verbose = $VERBOSE
                    $VERBOSE = nil

                    catch(:valid) do
                        eval(""BEGIN{throw :valid}\n#{str}"", b)
                    end

                    # Non comment lines that end in a comma or backslash are not complete.
                    (str =~ /^\s*#/ ? true : str !~ /[,\\]\s*\z/)

                    rescue SyntaxError => error
                        false
                    ensure
                        $VERBOSE = old_verbose
                end

                def get_binding(param)
                    return binding
                end

                clipStr = TkClipboard.get(nil)

                strBinding = get_binding(clipStr)

                ctrlStr = """"
                condStr = """"
                condFiltStr = """"
                keepBlank = true
                condCheck = true
                condCompEarly = false
                condMode = false
                ctrlCondComp = false
                ctrlStrComp = false
                condCount = 0
                condIndex = 0
                condArray = Array.new

                # noMsg is used in prototype tool used in developing this algorithm.
                noMsg = false

                condIfArray = [""if(!individual.married_joint.strip.empty?)"",
	                ""if(!individual.married_separate.strip.empty?)"",
	                ""if(!(individual.married_joint+individual.married_separate).strip.empty?)"",
	                ""if(!(individual.married_separate+individual.married_joint).strip.empty?)"",
	                ""if(!individual.business_name.strip.empty?)"",
	                ""if(!preparer.self_employed.strip.empty?)""]

                # TAKE OUT THIS NOMSG CHECK FOR RETRIEVER VERSION.
                #if (clipStr =~ /^NOMSG/)
                #    noMsg = true
                #    clipStr.sub!(/^NOMSG/, """")
                #end

                clipStr.each do |line|

	                lineNotEmpty = !line.gsub(/\s+/, '').empty?

	                if (keepBlank or lineNotEmpty)

		                keepBlank = lineNotEmpty
		                ctrlStr += line

		                ctrlStrComp = CompleteExpression?(ctrlStr, strBinding)

		                if (ctrlStrComp)

			                if (condMode)
				                ctrlCondComp = true
			                else
				                print ctrlStr + ""\x03""
				                condCheck = true
				                condCompEarly = false
				                ctrlStr = """"
			                end
		                end

		                if (!ctrlStrComp or ctrlCondComp)

			                unless (condCompEarly)

				                if (condMode)

					                condStr += line

					                if (CompleteExpression?(condStr, strBinding))

						                condArray[condIndex] = condStr
						                condIndex += 1
						                condStr = """"
						                condCheck = true

					                elsif (condCheck)

						                condStrFilt = condStr.gsub(/\s+/, '')
						
						                if (condIfArray.include?(condStrFilt))
			
							                condArray[condIndex] = condStr
							                condIndex += 1
							                condCount += 1
							                condStr = """"
						                else
							                if ([""end"",""else""].include?(condStrFilt))
			
								                condArray[condIndex] = condStr
								                condIndex += 1
								                condStr = """"

								                if (condStrFilt == ""end"")
			
									                condCount -= 1
			
									                if (condCount == 0)

										                if (ctrlCondComp)

											                print condArray.join(""\x03"") + ""\x03""
											                ctrlStr = """"
											                ctrlCondComp = false
											                condCompAck = true
										                else
											                condCompEarly = true
										                end
									                end
								                end
							                else
								                condCheck = false
							                end
						                end
					                end

					                if (ctrlCondComp or condCompEarly or condCompAck)

						                if (ctrlCondComp)
							                print ctrlStr + ""\x03""
							                ctrlStr = """"
							                ctrlCondComp = false
						                end

						                condArray.clear
						                condMode = false
						                condCompAck = false
						                condIndex = 0
						                condCount = 0
						                condStr = """"
					                end

				                elsif (condCheck)

					                if (condIfArray.include?(ctrlStr.gsub(/\s+/, '')))

						                condMode = true
						                condArray[condIndex] = ctrlStr
						                condIndex += 1
						                condCount += 1
					                else
						                condCheck = false
					                end
				                end
			                end
		                end
	                end
                end

                unless (ctrlStr.empty?)

                    # noMsg is used in prototype tool used in developing this algorithm.
	                unless (noMsg)

                        # Checking if someone made a bad selection, we will make it obvious.
	                    ctrlStr += (CompleteExpression?(ctrlStr, strBinding) ? """" : ""\nNOT COMPLETE EXPRESSION"")
                    end
	                print ctrlStr + ""\x03""
                end";
        }

        private void SetupGetTruncProcess()
        {
            _getTruncProcess = new Process();

            _getTruncProcess.StartInfo.UseShellExecute = false;
            _getTruncProcess.StartInfo.RedirectStandardOutput = true;
            _getTruncProcess.StartInfo.RedirectStandardInput = true;
            _getTruncProcess.StartInfo.CreateNoWindow = true;
            _getTruncProcess.StartInfo.FileName = _rubyExe;
        }

        public List<string> GetTruncationData()
        {
            /*
             * Runs the Ruby Interpreter via a Process(), utilizing the preset _getTruncScript
             * to see if the Ruby code which has been placed in the clipboard by the caller
             * is a complete expression.
            */
            _getTruncProcess.Start();
            _getTruncProcess.StandardInput.Write(_getTruncScript);
            _getTruncProcess.StandardInput.Flush();
            _getTruncProcess.StandardInput.Close();

            // To avoid deadlocks, always read the output stream first and then wait.
            string output = _getTruncProcess.StandardOutput.ReadToEnd();
            char etx = (char)3;
            List<string> truncList = output.Split(new char[] { etx }, StringSplitOptions.RemoveEmptyEntries).ToList();

            _getTruncProcess.WaitForExit();

            return truncList;
        }

        public List<string> GetRubyExpressions()
        {
            /*
             * Runs the Ruby Interpreter via a Process(), utilizing the preset _getExprScript
             * to generate a list of strings comprising complete Ruby expressions. Comment lines
             * and single empty lines are included.
            */
            Process p = new Process();
            p.StartInfo.UseShellExecute = false;
            p.StartInfo.RedirectStandardOutput = true;
            p.StartInfo.RedirectStandardInput = true;
            p.StartInfo.CreateNoWindow = true;
            p.StartInfo.FileName = _rubyExe;

            p.Start();

            p.StandardInput.Write(_getExprScript);
            p.StandardInput.Flush();
            p.StandardInput.Close();

            // To avoid deadlocks, always read the output stream first and then wait.
            // To avoid dreadlocks, steer clear of Jamaican culture, sorry Bob.
            string output = p.StandardOutput.ReadToEnd();
            char etx = (char)3;
            List<string> exprList = output.Split(new char[] { etx }, StringSplitOptions.RemoveEmptyEntries).ToList();

            p.WaitForExit();

            return exprList;
        }
    }
}